package com.sillycycle.bagleyd.barrel;

/*
 * @(#)BarrelCanvas.java
 *
 * Copyright 2004 - 2022  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Point;
import java.awt.Toolkit;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.IOException;
import java.net.URL;
import java.util.Random;
import javax.swing.JFrame;

/**
 * The <code>BarrelCanvas</code> class does all the graphics for
 * <code>BarrelFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

public class BarrelCanvas extends Canvas {
	private static final long serialVersionUID = 42L;
	public static final String NL = System.getProperty("line.separator");
	static final int MAX_ITERATIONS = 64;
	static final int MAX_SLICES = 16;

	static final int TOP = 0;
	static final int RIGHT = 1;
	static final int BOTTOM = 2;
	static final int LEFT = 3;
	static final int COORD = 4;
	static final int MAX_COLORS = BarrelInterface.MAX_FACES + 1;
	static final int PLUNGERS_NUMBER = 3;
	static final int NORMAL = 1;
	static final int DOUBLE = 2;
	static final int INSTANT = 3;
	static final String DATAFILE = "barrel.dat";
	static final String SYMBOL = ":";
	static final String PICTUREFILE = BarrelInterface.IMAGE;
	static final int PICTUREX = 512;
	static final int PICTUREY = 512;

	Color       foreground, background;
	Color       tileColor, borderColor;
	Color       currentForeground;
	Color[]     faceColor = new Color[MAX_COLORS];
	int[]       tileOfPosition = null;
	int[]       startPosition = null;
	int	 currentPosition = -1, currentSpace = -1;
	int	 currentDirection, lastDirection;
	int	 currentTile = -1, currentShift = -1;
	int	 nextPosition = -1, nextShift = -1;
	int	 columnPosition;
	int[]       spacePosition = new int[PLUNGERS_NUMBER];
	int[]       startSpace = new int[PLUNGERS_NUMBER];
	boolean     started, cheat, motion;
	boolean     practice, orient, mono, reverse, pairs;
	boolean     number = BarrelInterface.DEFAULT_NUMBER;
	int	 delay, numSlices, motionDiff, shiftDiff;
	int	 base;
	int	 lastX;
	int	 tiles, faces, tileFaces;
	Point       coreSize;
	Point       offset;
	Point       faceSize, tileSize, puzzleSize;
	Point       delta, puzzleOffset;

	JFrame      frame = null;
	int	 randomPaints = 0;
	boolean     firstPaint = false, resizePaint = false;
	boolean     movePaint = false, scanPaint = false, stackPaint = false;
	boolean     framePaint = false, allTilesPaint = false;
	boolean     selectPaint = false, motionPaint = false;
	boolean     releasePaint = false;
	boolean     mouseDown = false;
	Random      generator;
	BarrelStack undo = null, redo = null;
	BarrelSolve solve = null;
	String      stringSave = null;
	String[]    token = null;
	String      faceNames = null;
	Image       bufferedFrameImage = null;
	Image[]     bufferedStartImage = new Image[2];
	Graphics    bufferedFrameGraphics = null;
	Graphics[]  bufferedStartGraphics = new Graphics[2];

	@Override
	protected String paramString() {
		return super.paramString() +
		",orient=" + orient +
		",pairs=" + pairs +
		",practice=" + practice +
		",tiles=" + tiles +
		",faces=" + faces +
		",base=" + base +
		",mono=" + mono +
		",reverse=" + reverse +
		",names=" + faceNames +
		",color0=" + faceColor[0] +
		",color1=" + faceColor[1] +
		",color2=" + faceColor[2] +
		",color3=" + faceColor[3] +
		",color4=" + faceColor[4] +
		",color5=" + faceColor[5] +
		",color6=" + faceColor[6] +
		",color7=" + faceColor[7] +
		",tileColor=" + tileColor +
		",borderColor=" + borderColor +
		",foreground=" + foreground +
		",background=" + background +
		",delay=" + delay;
	}

	void checkTiles() {
		if (tiles < BarrelInterface.MIN_TILES) {
			System.out.println(tiles +
			" number of barrel in a row incorrect, must be at least " +
			BarrelInterface.MIN_TILES +
			", defaulting to " +
			BarrelInterface.DEFAULT_TILES);
			tiles = BarrelInterface.DEFAULT_TILES;
		}
		if (faces < BarrelInterface.MIN_FACES ||
				faces > BarrelInterface.MAX_FACES) {
			System.out.println(faces +
			" number of barrel in a row incorrect, must be at between " +
			BarrelInterface.MIN_FACES + " and " +
			BarrelInterface.MAX_FACES + "inclusive, defaulting to " +
			BarrelInterface.DEFAULT_FACES);
			tiles = BarrelInterface.DEFAULT_FACES;
		}
		if (base < BarrelInterface.MIN_BASE ||
				base > BarrelInterface.MAX_BASE) {
			System.out.println(base +
			" base incorrect, must be between " +
			BarrelInterface.MIN_BASE +
			" and " +
			BarrelInterface.MAX_BASE +
			" inclusive, defaulting to " +
			BarrelInterface.DEFAULT_BASE);
			base = BarrelInterface.DEFAULT_BASE;
		}
		if (delay < 0) {
			System.out.println(delay +
			" delay cannot be negative, taking absolute value");
			delay = -delay;
		}
	}

	void setColumnPosition(int x) {
		columnPosition = (x - delta.x / 2 - puzzleOffset.x) / offset.x;
		if (columnPosition < 0 || columnPosition >= tiles)
			columnPosition = -1;
	}

	boolean positionTile(int x) {
		setColumnPosition(x);
		return (columnPosition >= 0 && columnPosition < tiles);
	}

	int positionToTile(int x, int y) {
		int temp = faces * offset.y + delta.y - 1;
		int i = (x - delta.x / 2 - puzzleOffset.x) /
			offset.x;
		int j = ((y - delta.y / 2 - puzzleOffset.y + temp) %
			temp) / offset.y;

		if (i >= 0 && j >= 0 && i < tiles && j < faces)
			return (i + tiles * j);
		return -1;
	}

	int toColumn(int pos) {
		return (pos % tiles);
	}

	int toRow(int pos) {
		return (pos / tiles);
	}

	int cartesianX(int pos) {
		return toColumn(pos) * offset.x + delta.x + puzzleOffset.x;
	}

	int cartesianY(int pos) {
		return toRow(pos) * (offset.y + 1) + delta.y + puzzleOffset.y;
	}

	int tileNFromSpace(int plunger, int n, int direction) {
		int pos = spacePosition[plunger];

		if (direction == LEFT || direction == RIGHT)
			return (pos + ((direction == RIGHT) ? -n : n));
		/* direction == TOP || direction == BOTTOM */
		return (pos + (tiles * ((direction == BOTTOM) ? -n : n)));
	}

	boolean withinFrame(int x, int y, int dx, int dy) {
		return (x < dx + tileSize.x / 2 &&
			x > dx - tileSize.x / 2 &&
			tileSize.y * (x - dx) < tileSize.x * (y - dy) &&
			tileSize.y * (dx - x) < tileSize.x * (y - dy) &&
			tileSize.y * (x - dx + 2 * tileSize.x) >
				tileSize.x * (y - dy) &&
			tileSize.y * (dx - x + 2 * tileSize.x) >
				tileSize.x * (y - dy));
	}

	int solvedPosition(int tile) {
		if (tile >= tileFaces - faces + 1 &&
				tile <= tileFaces - faces + PLUNGERS_NUMBER)
			return tile - tileFaces + 1;
		return (tile + 1) % tileFaces;
	}

	boolean checkStarted() {
		return started;
	}

	boolean checkSolved() {
		int color, i, j, inc;

		if (tiles < 2)
			return true;
		if (orient) {
			for (j = 0; j < faces; j++) {
				color = (tileOfPosition[j * tiles + 1] - 1) /
					(tiles - 2);
				inc = (tileOfPosition[j * tiles + 1] - 1) %
					(tiles - 2);
				for (i = 1; i < tiles - 2; i++) {
					if (color != (tileOfPosition[j *
						tiles + i + 1] - 1) /
							(tiles - 2)) {
						return false;
					}
					if (inc + i != (tileOfPosition[j *
						tiles + i + 1] - 1) %
							(tiles - 2)) {
						return false;
					}
				}
				if (((color & 1) == 0) && ((tiles - 2) *
						faces + 1 + color / 2 !=
						tileOfPosition[j * tiles +
						tiles - 1])) {
					return false;
				}
			}
		} else {
			for (j = 0; j < faces; j++) {
				color = (tileOfPosition[j * tiles + 1] - 1) /
					(tiles - 2);
				for (i = 0; i < tiles - 3; i++) {
					if (color != (tileOfPosition[j *
						tiles + i + 2] - 1) /
							(tiles - 2)) {
						return false;
					}
				}
			}
		}
		started = false;
		return true;
	}

	int tilePicture(int tile) {
		int face, inc = 0;

		if (number)
			return tile;
		if (tile < 0)
			return -1;
		for (face = 0; face < faces; face++) {
			inc += tiles - 2;
			if (tile < inc)
				return tile + 2 * face + 1;
		}
		for (face = 0; face < PLUNGERS_NUMBER; face++) {
			if (tile == inc++)
				return (face * 2 + 1) * tiles - 1;
		}
		for (face = 0; face < faces - PLUNGERS_NUMBER; face++) {
			if (tile == inc++)
				return (2 * face + 1) * tiles;
			if (tile == inc++)
				return (2 * face + 2) * tiles - 1;
		}
		return 0;
	}

	static void fillCircle(Graphics g, int diameter, int x, int y) {
		if (diameter <= 0)
			return;
		g.fillOval(x - diameter / 2, y - diameter / 2,
			diameter, diameter);
	}

	void drawBall(Graphics g, int pos, int offsetX, int offsetY) {
		int dx, dy, diameter;

		diameter = java.lang.Math.min(offset.x, offset.y) * 3 / 4 - 3;
		dx = cartesianX(pos) + tileSize.x / 2 - offsetX;
		dy = cartesianY(pos) + tileSize.y / 2 - offsetY;
		fillCircle(g, diameter, dx, dy);
	}

	void copyImage(Graphics g, Image i,
			boolean scale,
			int dX, int dY, int sX, int sY,
			int width, int height) {
		if (scale) {
			g.drawImage(i, dX, dY, dX + width, dY + height,
				sX * PICTUREX / coreSize.x,
				sY * PICTUREY / coreSize.y,
				(sX + width) * PICTUREX / coreSize.x,
				(sY + height) * PICTUREY / coreSize.y, this);
		} else {
			g.drawImage(i, dX, dY, dX + width, dY + height,
				sX, sY, sX + width, sY + height, this);
		}
	}

	static void drawShadow(Graphics g, int startX, int startY,
			int sizeX, int sizeY) {
		g.drawLine(startX, startY, startX + sizeX, startY);
		g.drawLine(startX, startY, startX, startY + sizeY);
	}

	void drawTile(Graphics g, int pos,
			boolean blank, boolean erase,
			int pressedOffset, int offsetX, int offsetY) {
		int dx, dy, sx, sy;
		int tile = tilePicture((tileOfPosition[pos] + tileFaces - 1) %
			tileFaces);

		dx = cartesianX(pos) + offsetX + pressedOffset;
		dy = cartesianY(pos) + offsetY + pressedOffset;
		sx = cartesianX(tile);
		sy = cartesianY(tile);
		if (pressedOffset == 0) {
			g.setColor(background);
			g.drawLine(dx + tileSize.x, dy,
				dx + tileSize.x, dy + tileSize.y);
			g.drawLine(dx, dy + tileSize.y,
				dx + tileSize.x, dy + tileSize.y);
		}
		if (blank) {
			g.setColor((erase) ? background : tileColor);
			g.fillRect(dx, dy, tileSize.x, tileSize.y);
			return;
		}
		copyImage(g, bufferedStartImage[pressedOffset], !number,
			dx, dy, sx, sy, tileSize.x, tileSize.y);
	}

	void drawTile(int pos, boolean blank, boolean erase,
			int pressedOffset, int offsetX, int offsetY) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			drawTile(g, pos, blank, erase,
				pressedOffset, offsetX, offsetY);
		} finally {
			g.dispose();
		}
	}

	void drawBufferedTile(Graphics g, int pos, int pressedOffset) {
		int dx, dy, tile = solvedPosition(pos);
		int i, digitX = 0, digitY = -1, fontSize = 0;
		String buf = "";
		Color ballColor, fillColor, lineColor;

		if (g == null)
			return;
		dx = cartesianX(pos) + pressedOffset;
		dy = cartesianY(pos) + pressedOffset;
		if ((pressedOffset != 0) ^ reverse) {
			fillColor = borderColor;
			lineColor = tileColor;
		} else {
			fillColor = tileColor;
			lineColor = borderColor;
		}
		if (pressedOffset != 0) {
			g.setColor(fillColor.darker());
			drawShadow(g,
				dx - pressedOffset, dy - pressedOffset,
				tileSize.x - 1, tileSize.y - 1);
		}
		g.setColor(fillColor);
		if (mono || tile <= 0) {
			ballColor = lineColor;
		} else {
			int face = ((tile - 1) / (tiles - 2)) % MAX_COLORS;
			ballColor = faceColor[face];
		}
		g.fill3DRect(dx, dy, tileSize.x, tileSize.y,
			pressedOffset == 0);
		if (tile > 0 && tile <= ((tiles - 2) * faces + PLUNGERS_NUMBER)) {
			g.setColor(ballColor);
			drawBall(g, pos, pressedOffset, pressedOffset);
			if (mono || orient) {
				g.setColor(tileColor);
				fontSize = java.lang.Math.min(tileSize.x,
					tileSize.y) / 2;
				g.setFont(new Font("arial", Font.PLAIN,
					fontSize));
				digitY = fontSize / 3;
			}
			if (orient) {
				i = tile;
				while (i >= 1) {
					i /= base;
					digitX += -fontSize / 3;
				}
				if (mono && (tile != 0 &&
					((tile != tileFaces - 1 &&
					tiles > 1) || tiles > 2)))
				  buf = Integer.toString(tile, base).toUpperCase() +
					faceNames.charAt(((tile - 1) /
					(tiles - 2)) % MAX_COLORS);
				else
				  buf = Integer.toString(tile, base).toUpperCase();
			} else if (mono && (tile != 0 &&
				  ((tile != tileFaces - 1 && tiles > 1) ||
				  tiles > 2)))
				buf = Character.toString(faceNames.charAt(((tile - 1) /
					(tiles - 2)) % MAX_COLORS));
			if (mono && (tile != 0 &&
			    ((tile != tileFaces - 1 && tiles > 1) ||
			    tiles > 2)))
				digitX += -fontSize / 3;
			if (mono && (tile != 0 &&
			    ((tile != tileFaces - 1 && tiles > 1) ||
			    tiles > 2)) || orient)
			    g.drawString(buf,
				dx + tileSize.x / 2 + digitX +
				tileSize.x / 25,
				dy + tileSize.y / 2 + digitY);
		}
	}

	void drawAllTiles() {
		for (int k = 0; k < tileFaces; k++) {
			drawTile(k, (tileOfPosition[k] < 0),
				(tileOfPosition[k] < 0), 0, 0, 0);
		}
	}

	void drawAllBufferedTiles() {
		if (number) {
			int k, l;

			for (k = 0; k < tileFaces; k++) {
				for (l = 0; l < 2; l++) {
					drawBufferedTile(
						bufferedStartGraphics[l],
						k, l);
				}
			}
		} else {
			URL url = BarrelFrame.class.getResource("/" + PICTUREFILE);

			if (url == null)
				bufferedStartImage[1] = null;
			else
				bufferedStartImage[1] = Toolkit.getDefaultToolkit().getImage(url);
			bufferedStartImage[0] = bufferedStartImage[1];
		}
	}

	void drawBuffer(Graphics g, Image i) {
		g.drawImage(i, 0, 0, coreSize.x, coreSize.y, this);
	}

	void drawBuffer(Image i) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			drawBuffer(g, i);
		} finally {
			g.dispose();
		}
	}

	void animateSlide(int numOfTiles, int dir,
			int fast, boolean logMoves) {
		Graphics g = getGraphics();
		int inc, aTile, plunger;
		int[] fillBeginPos = new int[PLUNGERS_NUMBER];
		int fillBeginI, gapI, moveI;
		int dx, dy, pos, posNext;
		int ix, iy;
		int numTiles = numOfTiles;

		if (g == null)
			return;
		try {
			for (plunger = 0; plunger < PLUNGERS_NUMBER; plunger++) {
			  fillBeginPos[plunger] =
			    tileNFromSpace(plunger, numTiles, dir);
			}
			fillBeginI = cartesianX(fillBeginPos[0]);
			gapI = tileSize.x * fast / numSlices;
			moveI = tileSize.x + delta.x;
			if (gapI == 0)
			  gapI++;
			if (numTiles < 0)
			  numTiles = -numTiles;
			for (inc = 0; inc < moveI + gapI; inc += gapI) {
			  if (inc > moveI)
			    inc = moveI;
			  for (plunger = 0; plunger < PLUNGERS_NUMBER; plunger++) {
			    for (aTile = 0; aTile < numTiles; aTile++) {
			      posNext = tileNFromSpace(plunger, aTile + 1,
				dir);
			      if (logMoves && inc == 0 &&
				  plunger == 0 && aTile == 0) {
				//set direction for undo
				((BarrelFrame) frame).callback(
				  BarrelInterface.ACTION_MOVED);
				undo.setMove(dir, posNext, 0);
				redo.flushMoves();
			      }
			      /* Calculate deltas */
			      dx = cartesianX(posNext);
			      dy = cartesianY(posNext);
			      ix = ((dir == RIGHT) ? inc : -inc);
			      iy = 0;
			      drawTile(posNext, false, false, 0, ix, iy);
			      /* Erase old slivers */
			      ix += dx;
			      iy += dy;
			      g.setColor(background);
			      if (aTile < numTiles - 1) {
				if (dir == RIGHT) {
				  g.fillRect(ix - gapI, iy,
				    gapI, tileSize.y);
				} else if (dir == LEFT) {
				  g.fillRect(ix + tileSize.x, iy,
				    gapI, tileSize.y);
				}
			      } else {
				if (dir == RIGHT) {
				  g.fillRect(fillBeginI, iy,
				    ix - fillBeginI, tileSize.y);
				} else if (dir == LEFT) {
				  g.fillRect(ix + tileSize.x, iy,
				    fillBeginI - ix, tileSize.y);
				}
			      }
			    }
			    try {
			      Thread.sleep(delay / fast);
			    } catch (Exception e) {
			      // ignore
			    }
			  }
			}
			for (plunger = 0; plunger < PLUNGERS_NUMBER; plunger++) {
			  pos = spacePosition[plunger];
			  for (aTile = 0; aTile < numTiles; aTile++) {
			    posNext = tileNFromSpace(plunger, aTile + 1, dir);
			    tileOfPosition[pos] = tileOfPosition[posNext];
			    pos = posNext;
			  }
			  spacePosition[plunger] = fillBeginPos[plunger];
			  tileOfPosition[fillBeginPos[plunger]] = -PLUNGERS_NUMBER +
			    plunger;
			}
		} finally {
			g.dispose();
		}
	}

	void drawAllTilesForColumns(Graphics g, int start, int finish,
			int dx, int dy) {
		int i, j, k;

		for (i = start; i <= finish; i++) {
			for (j = 0; j < faces; j++) {
				k = j * tiles + i;
				if (tileOfPosition[k] > 0)
					drawTile(g, k,
						(tileOfPosition[k] < 0),
						(tileOfPosition[k] < 0),
						0, -dx, -dy);
			}
		}
	}

	void animateRotate(int start, int finish, int direction, int fast) {
		Image columnImage;
		Graphics columnGraphics;
		Graphics g = getGraphics();
		int size = finish - start + 1;
		int moveI, gapJ, moveJ, inc, i, animateOffset = 0;

		if (g == null)
			return;
		try {
			moveI = tileSize.x + delta.x;
			gapJ = tileSize.y * fast / numSlices;
			moveJ = tileSize.y + delta.y + 1;
			if (gapJ == 0)
				gapJ++;
			columnImage = createImage(size * moveI,
				puzzleSize.y);
			columnGraphics = columnImage.getGraphics();
			copyImage(columnGraphics, bufferedFrameImage,
				false, 0, 0,
				start * moveI + puzzleOffset.x + 1,
				puzzleOffset.y,
				size * moveI, puzzleSize.y);
			drawAllTilesForColumns(columnGraphics, start, finish,
				(start * moveI + puzzleOffset.x + 1),
				puzzleOffset.y);
			for (inc = 0; inc < moveJ + gapJ; inc += gapJ) {
				if (inc > moveJ)
					inc = moveJ;
				if (direction == TOP) {
					i = inc;
				} else {
					i = puzzleSize.y - inc + 1;
					animateOffset = 3;
				}
				copyImage(g, columnImage,
					false,
					start * moveI + puzzleOffset.x + 1,
					puzzleOffset.y + animateOffset,
					0, i,
					size * moveI,
					puzzleSize.y - i - 1);
				copyImage(g, columnImage,
					false,
					start * moveI + puzzleOffset.x + 1,
					puzzleSize.y - i + puzzleOffset.y -
					delta.y - 1 + animateOffset,
					0, 0,
					size * moveI,
					i - delta.y - 1);
				try {
					Thread.sleep(delay / fast);
				} catch (InterruptedException e) {
					// ignore
				}
			}
		} finally {
			g.dispose();
		}
	}

	void drawFrame(Graphics g, boolean show) {
		int offsetX = puzzleOffset.x;
		int offsetY = puzzleOffset.y;

		if (g == null)
			return;
		if (show) {
			int sumX, sumY, k;

			g.setColor(borderColor);
			g.fillRect(0, 0, coreSize.x, offsetY);
			g.fillRect(0, coreSize.y - offsetY,
				coreSize.x, offsetY);
			g.fillRect(0, 0, offsetX, coreSize.y);
			g.fillRect(coreSize.x - offsetX, 0,
				offsetX, coreSize.y);
			g.setColor(currentForeground);
			sumX = tiles * offset.x + 1;
			sumY = offset.y + 1;
			for (k = 0; k < faces; k++) {
				g.fillRect(offsetX, offsetY, 1, sumY);
				g.fillRect(offsetX, offsetY, sumX, 1);
				g.fillRect(sumX + offsetX, offsetY,
					1, sumY + 1);
				offsetY += sumY;
				g.fillRect(offsetX, offsetY, sumX + 1, 1);
			}
		} else {
			/*
			g.setColor(Color.white);
			g.fillRect(0, 0, coreSize.x, offsetY);
			g.fillRect(0, coreSize.y - offsetY - 1,
				coreSize.x, offsetY + 1);
			g.fillRect(0, 0, offsetX, coreSize.y);
			g.fillRect(coreSize.x - offsetX - 1, 0,
				offsetX + 1, coreSize.y);
			*/
			g.setColor(background);
			g.fillRect(offsetX + 1, offsetY + 1,
				coreSize.x - 2 - 2 * offsetX,
				coreSize.y - 2 - 2 * offsetY);
		}
	}

	void drawFrame(boolean show) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			drawFrame(g, show);
		} finally {
			g.dispose();
		}
	}

	void moveNoTiles() {
		((BarrelFrame) frame).callback(
			BarrelInterface.ACTION_IGNORE);
	}

	void moveTiles(int fast) {
		int pos, posNext, plunger, aTile, dir;

		if (fast != INSTANT && delay > 0) {
			if (spacePosition[0] % tiles == 0) {
				animateSlide(tiles - 1, LEFT, fast, false);
			} else {
				animateSlide(tiles - 1, RIGHT, fast, false);
			}
		} else {
			dir = (spacePosition[0] % tiles == 0) ? LEFT : RIGHT;
			for (plunger = 0; plunger < PLUNGERS_NUMBER; plunger++) {
				pos = spacePosition[plunger];
				for (aTile = 0; aTile < tiles - 1; aTile++) {
					posNext = tileNFromSpace(plunger, 1, dir);
					tileOfPosition[pos] = tileOfPosition[posNext];
					drawTile(pos, false, false, 0, 0, 0);
					pos = posNext;
					spacePosition[plunger] = pos;
				}
				tileOfPosition[pos] = -PLUNGERS_NUMBER + plunger;
				drawTile(pos, true, true, 0, 0, 0);
			}
		}
		if (((BarrelFrame) frame).getToggleSound())
			((BarrelFrame) frame).playBumpAudio();
	}

	void rotateTiles(int start, int finish, int direction, int fast) {
		int newR, i, k, tempTile, currTile, pos, newPos;
		int r = 0, plunger;
		boolean animate = (fast != INSTANT && delay > 0);

		if (animate)
			animateRotate(start, finish, direction, fast);
		for (i = start; i <= finish; i++) {
			currTile = tileOfPosition[r * tiles + i];
			for (k = 1; k <= faces; k++) {
				newR = (((direction == TOP) ?
					(faces - 1) : 1) + r) % faces;
				pos = r * tiles + i;
				newPos = newR * tiles + i;
				tempTile = tileOfPosition[newPos];
				tileOfPosition[newPos] = currTile;
				currTile = tempTile;
				for (plunger = 0; plunger < PLUNGERS_NUMBER;
					plunger++) {
					if (spacePosition[plunger] == pos) {
						drawTile(newPos, true, true,
							0, 0, 0);
						break;
					}
				}
				if (plunger == PLUNGERS_NUMBER)
					drawTile(newPos, false, false,
						0, 0, 0);
				r = newR;
			}
		}
		if (start == 0) {
			for (plunger = 0; plunger < PLUNGERS_NUMBER; plunger++) {
				i = (direction == TOP) ? faces - 1 : 1;
				spacePosition[plunger] =
					(spacePosition[plunger] + i *
					tiles) % tileFaces;
			}
		}
		/* No longer needed to correct final image. */
		/*if (animate) {
			Graphics g = getGraphics();

			if (g == null)
				return;
			try {
				drawFrame(g, true);
				drawAllTilesForColumns(g, start, finish, 0, 0);
			} finally {
				g.dispose();
			}
		}*/
		if (((BarrelFrame) frame).getToggleSound())
			((BarrelFrame) frame).playMoveAudio();
	}

	boolean moveTilesDir(int direction, int tile, boolean all, int fast) {
		int start, finish;

		start = tile;
		finish = tile;
		if (all) {
			start = 0;
			finish = tiles - 1;
		} else if (pairs) {
			start = 1 + ((tile - 1) / 2) * 2;
			finish = start + 1;
		}
		switch (direction) {
		case TOP:
			if (!all && (tile == 0 || tile == tiles - 1)) {
				return false;
			}
			rotateTiles(start, finish, TOP, fast);
			return true;
		case RIGHT:
			if (toColumn(spacePosition[0]) > 0) {
				moveTiles(fast);
				return true;
			}
			break;
		case BOTTOM:
			if (!all && (tile == 0 || tile == tiles - 1)) {
				return false;
			}
			rotateTiles(start, finish, BOTTOM, fast);
			return true;
		case LEFT:
			if (toColumn(spacePosition[0]) < tiles - 1) {
				moveTiles(fast);
				return true;
			}
			break;
		default:
			System.out.println("moveTilesDir: direction " +
				direction);
		}
		return false;
	}

	boolean checkmoveTilesDir(int direction) {
		switch (direction) {
		case TOP:
			return true;
		case RIGHT:
			if (toColumn(spacePosition[0]) > 0)
				return true;
			return false;
		case BOTTOM:
			return true;
		case LEFT:
			if (toColumn(spacePosition[0]) < tiles - 1)
				return true;
			return false;
		default:
			System.out.println("checkmoveTilesDir: direction " +
				direction);
		}
		return false;
	}

	boolean movePuzzle(int direction, int tile, boolean shift,
			boolean motionFlag, int fast) {
		if (shift && (direction == TOP || direction == BOTTOM)) {
			moveShiftCb(direction, fast);
			return true;
		} else if (moveTilesDir(direction, tile, false, fast)) {
			started = true;
			if (!motionFlag) {
				((BarrelFrame) frame).callback(
					BarrelInterface.ACTION_MOVED);
				undo.setMove(direction, tile, 0);
				redo.flushMoves();
			}
			return true;
		}
		return false;
	}

	void diffMove(int diff, int j, boolean shift) {
		int k;

		if (diff > faces / 2 || (diff == (faces + 1) / 2 &&
				j > currentPosition / tiles))
			for (k = 0; k < faces - diff; k++)
				movePuzzle(BOTTOM, currentPosition % tiles,
					shift, true, NORMAL);
		else
			for (k = 0; k < diff; k++)
				movePuzzle(TOP, currentPosition % tiles,
					shift, true, NORMAL);
	}

	void moveRecord(int direction, int tile, boolean shift) {
		if (shift && (direction == TOP || direction == BOTTOM)) {
			undo.setMove(direction, tile, 1);
		} else {
			((BarrelFrame) frame).callback(
				BarrelInterface.ACTION_MOVED);
			undo.setMove(direction, tile, 0);
		}
		redo.flushMoves();
	}

	void diffRecord(int diff, int j, boolean shift) {
		int k;

		currentTile = toColumn(currentPosition);
		if (diff > faces / 2 || (diff == (faces + 1) / 2 &&
				j > currentPosition / tiles))
			for (k = 0; k < faces - diff; k++)
				moveRecord(BOTTOM,
					currentTile, shift);
		else
			for (k = 0; k < diff; k++)
				moveRecord(TOP,
					currentTile, shift);
	}

	void discreteSlide(int pos) {
		int dir = ((pos <= 0) ? RIGHT : LEFT);
		int oldPos = (pos <= 0) ? 0 : tiles - 1;

		moveTiles(NORMAL);
		((BarrelFrame) frame).callback(
			BarrelInterface.ACTION_MOVED);
		undo.setMove(dir, oldPos, 0);
		redo.flushMoves();
	}

	void slidePuzzle(int direction) {
		if (checkSolved() && !practice) {
			moveNoTiles();
			return;
		}
		if (!movePuzzle(direction, 0, false, false, NORMAL)) {
			((BarrelFrame) frame).callback(
				BarrelInterface.ACTION_BLOCKED);
			return;
		}
		if (checkSolved()) {
			((BarrelFrame) frame).callback(
				BarrelInterface.ACTION_SOLVED);
		}
	}

	void rotatePuzzle(int direction, int tile) {
		if (checkSolved() && !practice) {
			moveNoTiles();
			return;
		}
		movePuzzle(direction, tile, false, false, NORMAL);
		if (checkSolved()) {
			((BarrelFrame) frame).callback(
				BarrelInterface.ACTION_SOLVED);
		}
	}

	void movePuzzleInput(int x, int direction, boolean shift) {
		if (mouseDown || randomPaints != 0)
			return;
		if (stackPaint) {
			repaint();
			return;
		}
		if (shift && (direction == TOP || direction == BOTTOM)) {
			moveShiftCb(direction, NORMAL);
			undo.setMove(direction, 0, 1);
			redo.flushMoves();
		} else if (checkSolved() && !practice) {
			moveNoTiles();
		} else if (direction == LEFT || direction == RIGHT) {
			slidePuzzle(direction);
			if (checkSolved()) {
				((BarrelFrame) frame).callback(
					BarrelInterface.ACTION_SOLVED);
			}
		} else {
			if (!positionTile(x))
				return;
			rotatePuzzle(direction, columnPosition);
			if (checkSolved()) {
				((BarrelFrame) frame).callback(
					BarrelInterface.ACTION_SOLVED);
			}
		}
	}

	void controlPuzzle(int direction) {
		int reason = BarrelInterface.ACTION_IGNORE;
		switch (direction) {
		case TOP:
			if (faces >= BarrelInterface.MAX_FACES)
				return;
			reason = BarrelInterface.ACTION_INCY;
			break;
		case RIGHT:
			reason = BarrelInterface.ACTION_INCX;
			break;
		case BOTTOM:
			if (faces <= BarrelInterface.MIN_FACES)
				return;
			reason = BarrelInterface.ACTION_DECY;
			break;
		case LEFT:
			if (tiles <= BarrelInterface.MIN_TILES)
				return;
			reason = BarrelInterface.ACTION_DECX;
			break;
		default:
			System.out.println("ControlBarrel: direction " +
				direction);
		}
		((BarrelFrame) frame).callback(reason);
	}

	void motionTiles() {
		if (nextShift == 0 && !practice && checkSolved()) {
			moveNoTiles();
			return;
		}
		if (toRow(nextPosition) != toRow(currentPosition)) {
			int j = nextPosition / tiles;
			int diff;

			if (nextShift == 1 && (toColumn(currentPosition) == 0 ||
					toColumn(currentPosition) == tiles - 1)) {
				currentPosition = -1;
				return;
			}
			diff = (toRow(currentPosition) -
				toRow(nextPosition) + faces) % faces;
			diffMove(diff, j, nextShift != 0);
			if (nextShift == 0 && checkSolved()) {
				((BarrelFrame) frame).callback(
					BarrelInterface.ACTION_SOLVED);
			}
			motion = true;
			if (nextShift == 0)
				motionDiff = (diff + motionDiff) % faces;
			else
				shiftDiff = (diff + shiftDiff) % faces;
		}
	}

	void selectTiles() {
		int diff;

		if (nextShift == 0 && !practice && checkSolved()) {
			moveNoTiles();
			return;
		}
		if (toRow(nextPosition) == toRow(currentPosition) && !motion) {
			if (nextShift == 1 && checkSolved()) {
				moveNoTiles();
				return;
			}
			if ((spacePosition[0] % tiles) == 0)
				diff = tiles - 1;
			else
				diff = -tiles + 1;
			selectSlideTiles(diff);
			currentPosition = -1;
			started = true;
			if (checkSolved()) {
				((BarrelFrame) frame).callback(
					BarrelInterface.ACTION_SOLVED);
			}
		} else {
			int j = nextPosition / tiles;

			diff = (toRow(currentPosition) -
				toRow(nextPosition) + faces) % faces;
			diffMove(diff, j, nextShift != 0);
			if (nextShift == 0)
				motionDiff = (diff + motionDiff) % faces;
			else
				shiftDiff = (diff + shiftDiff) % faces;
			diffRecord(motionDiff, j, false);
			diffRecord(shiftDiff, j, true);
			if (nextShift == 0 && checkSolved()) {
				((BarrelFrame) frame).callback(
					BarrelInterface.ACTION_SOLVED);
			}
			currentPosition = nextPosition;
			motionDiff = 0;
			shiftDiff = 0;
		}
	}

	void selectSlideTiles(int pos) {
		if (pos < 0) {
			if (delay > 0) {
				animateSlide(-pos, RIGHT, NORMAL, true);
				if (((BarrelFrame) frame).getToggleSound())
					((BarrelFrame) frame).playBumpAudio();
			} else {
				discreteSlide(pos);
			}
		} else if (pos > 0) {
			if (delay > 0) {
				animateSlide(pos, LEFT, NORMAL, true);
				if (((BarrelFrame) frame).getToggleSound())
					((BarrelFrame) frame).playBumpAudio();
			} else {
				discreteSlide(pos);
			}
		}
	}

	void moveShiftCb(int direction, int fast) {
		moveTilesDir(direction, 0, true, fast);
		((BarrelFrame) frame).callback(
			BarrelInterface.ACTION_CONTROL);
	}

	void resetTiles() {
		int i, inc;

		solve.setFlag(false);
		tileFaces = tiles * faces;
		tileOfPosition = new int[tileFaces];
		startPosition = new int[tileFaces];
		inc = 0;
		for (i = 0; i < tileFaces; i++) {
			if (i % tiles != 0 && i % tiles != tiles - 1) {
				inc++;
				tileOfPosition[i] = inc;
			} else {
				tileOfPosition[i] = 0;
			}
		}
		for (i = 0; i < PLUNGERS_NUMBER; i++) {
			spacePosition[i] = 2 * i * tiles;
			tileOfPosition[2 * i * tiles] = -PLUNGERS_NUMBER + i;
			inc++;
			tileOfPosition[(2 * i + 1) * tiles - 1] = inc;
		}
		inc++;
		for (i = 0; i < faces - PLUNGERS_NUMBER; i++) {
			tileOfPosition[(2 * i + 1) * tiles] = inc++;
			tileOfPosition[(2 * i + 2) * tiles - 1] = inc++;
		}
		undo.flushMoves();
		redo.flushMoves();
		getStartPosition();
		currentPosition = -1;
		currentDirection = -1;
		started = false;
		cheat = false;
		randomPaints = 0;
	}

	void practiceTiles() {
		((BarrelFrame) frame).callback(BarrelInterface.ACTION_PRACTICE);
	}

	void randomizeTiles() {
		if (practice)
			practiceTiles();
		/* Now move the spaces around randomly */
		if (tiles > 1 || faces > 1) {
			((BarrelFrame) frame).callback(
				BarrelInterface.ACTION_RESET);
			lastDirection = -1;
			randomPaints = java.lang.Math.min(tileFaces,
				MAX_ITERATIONS) + generator.nextInt(4);
			repaint();
		} else if (checkSolved())
			((BarrelFrame) frame).callback(
				BarrelInterface.ACTION_SOLVED);
	}

	void randomizingTiles() {
		currentDirection = generator.nextInt(COORD);
		if ((currentDirection & 1) == 1) {
			if (spacePosition[0] % tiles == 0)
				currentDirection = LEFT;
			else
				currentDirection = RIGHT;
		}
		if ((currentDirection + COORD / 2) % COORD != lastDirection) {
			if ((currentDirection & 1) == 1) {
				movePuzzle(currentDirection, tiles - 1,
					false, false, INSTANT);
			} else {
				movePuzzle(currentDirection,
					generator.nextInt(tiles - 2) + 1,
					false, false, INSTANT);
			}
			lastDirection = currentDirection;
			currentDirection = -1;
			randomPaints--;
			try {
				Thread.sleep(delay);
			} catch (InterruptedException e) {
				// ignore
			}
		}
		if (randomPaints > 0) {
			repaint();
		} else {
			undo.flushMoves();
			redo.flushMoves();
			getStartPosition();
			((BarrelFrame) frame).callback(
				BarrelInterface.ACTION_RANDOMIZE);
			started = false;
			currentPosition = -1;
			if (checkSolved())
				((BarrelFrame) frame).callback(
					BarrelInterface.ACTION_SOLVED);
		}
	}

	public boolean getNumber() {
		return number;
	}

	public int getDelay() {
		return delay;
	}

	public void setDelay(int delay) {
		this.delay = delay;
		numSlices = ((delay < MAX_SLICES) ? delay + 1 : MAX_SLICES);
	}

	public boolean getCheat() {
		return cheat;
	}

	boolean setPuzzleOrient(boolean newValue) {
		if (orient != newValue) {
			orient = newValue;
			resetTiles();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	boolean setPuzzlePairs(boolean newValue) {
		if (pairs != newValue) {
			pairs = newValue;
			resetTiles();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	boolean setPuzzlePractice(boolean newValue) {
		if (practice != newValue) {
			practice = newValue;
			resetTiles();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	boolean setPuzzleTiles(int newValue) {
		if (tiles != newValue && newValue >= BarrelInterface.MIN_TILES) {
			tiles = newValue;
			resetTiles();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	boolean setPuzzleFaces(int newValue) {
		if (faces != newValue && newValue >= BarrelInterface.MIN_FACES) {
			faces = newValue;
			resetTiles();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	public void speedUpPuzzle() {
		delay -= 10;
		if (delay < 0)
			delay = 0;
		((BarrelFrame) frame).callback(BarrelInterface.ACTION_SPEED);
	}

	public void slowDownPuzzle() {
		delay += 10;
		((BarrelFrame) frame).callback(BarrelInterface.ACTION_SPEED);
	}

	void showMessage(String msg) {
		((BarrelFrame) frame).showMessage(msg);
	}

	void resizePuzzle() {
		coreSize = new Point(getSize().width, getSize().height);
		delta = new Point(1, 1);
		offset = new Point(java.lang.Math.max((coreSize.x - 2) /
			tiles, 0),
			java.lang.Math.max(coreSize.y / faces - 2, 0));
		faceSize = new Point(offset.x * tiles + delta.x + 2,
			offset.y + delta.y + 2);
		puzzleSize = new Point(faceSize.x,
			(faceSize.y - 2) * faces + 2);
		puzzleOffset = new Point((coreSize.x - puzzleSize.x + 2) / 2,
			(coreSize.y - puzzleSize.y + 2) / 2);
		tileSize = new Point(java.lang.Math.max(offset.x - delta.x, 0),
			java.lang.Math.max(offset.y - delta.y, 0));
		bufferedFrameImage = createImage(coreSize.x, coreSize.y);
		bufferedFrameGraphics = bufferedFrameImage.getGraphics();
		if (number) {
			bufferedStartImage[0] =
				createImage(coreSize.x, coreSize.y);
			bufferedStartImage[1] =
				createImage(coreSize.x, coreSize.y);
		} else {
			bufferedStartImage[0] =
				createImage(PICTUREX, PICTUREY);
			bufferedStartImage[1] =
				createImage(PICTUREX, PICTUREY);
		}
		bufferedStartGraphics[0] = bufferedStartImage[0].getGraphics();
		bufferedStartGraphics[1] = bufferedStartImage[1].getGraphics();
	}

	public void initializePuzzle() {
		checkTiles();
		undo = new BarrelStack();
		redo = new BarrelStack();
		solve = new BarrelSolve(frame, this);
		resetTiles();
		generator = new Random(System.nanoTime());
	}

	void exposePuzzle() {
		drawFrame(bufferedFrameGraphics, false);
		drawFrame(bufferedFrameGraphics, true);
		drawBuffer(bufferedStartGraphics[0], bufferedFrameImage);
		drawBuffer(bufferedStartGraphics[1], bufferedFrameImage);
		drawAllBufferedTiles();
		drawBuffer(bufferedFrameImage);
		drawAllTiles();
		solve.init();
	}

	void selectPuzzle(boolean shift, int x, int y) {
		motion = false;
		motionDiff = 0;
		shiftDiff = 0;
		currentPosition = positionToTile(x, y);
		if (currentPosition >= 0) {
			if (!shift && !practice && checkSolved()) {
				moveNoTiles();
				currentPosition = -1;
				return;
			}
			selectPaint = true;
			repaint();
		} else
			currentPosition = -1;
	}

	void motionPuzzle(boolean shift, int x, int y) {
		if (currentPosition == -1)
			return;
		if (!shift && !practice && checkSolved()) {
			moveNoTiles();
			currentPosition = -1;
			return;
		}
		nextShift = (shift) ? 1 : 0;
		nextPosition = positionToTile(x, y);
		motionPaint = true;
		repaint();
	}

	void releasePuzzle(boolean shift, int x, int y) {
		if (currentPosition == -1)
			return;
		nextPosition = positionToTile(x, y);
		if (!shift && (currentPosition % tiles == 0 ||
				currentPosition % tiles == tiles - 1) &&
				currentPosition != nextPosition) {
			currentSpace = -1;
		} else {
			currentSpace = 0;
		}
		nextShift = (shift) ? 1 : 0;
		releasePaint = true;
		repaint();
	}

	public void clearPuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		resetTiles();
		drawAllTiles();
		allTilesPaint = true;
		repaint();
		((BarrelFrame) frame).callback(
			BarrelInterface.ACTION_RESET);
	}

	void randomizePuzzle() {
		if (!mouseDown)
			randomizeTiles();
	}

	void randomizePuzzleWithQuery() {
		if (!started)
			randomizePuzzle();
	}

	public void getPuzzle() {
		this.stringSave = ((BarrelFrame) frame).readTextArea();
		if (this.stringSave == null || this.stringSave.equals("")) {
			showMessage("Empty string to read");
			return;
		}
		getPuzzleString("buffer");
	}

	public void getPuzzleFile() {
		undo.flushMoves();
		redo.flushMoves();
		getStartPosition();
		try (BufferedReader br = new BufferedReader(
				new FileReader(DATAFILE))) {
			StringBuilder puzzleData = new StringBuilder();
			String lineData;

			while ((lineData = br.readLine()) != null)
				puzzleData.append(lineData).append(NL);
			this.stringSave = puzzleData.toString();
			if (this.stringSave == null || this.stringSave.equals("")) {
				showMessage("Empty file " + DATAFILE);
				return;
			}
			getPuzzleString(DATAFILE);
		} catch (IOException e) {
			//e.printStackTrace();
			showMessage("Cannot read from " + DATAFILE);
		}
	}

	public void getPuzzleString(String source) {
		try {
			int i, tilesValue, facesValue, moves;
			boolean orientValue, pairsValue, practiceValue;
			String symbol = SYMBOL;
			String newline = NL;
			int line = 0;

			token = this.stringSave.split(newline);
			String lineData = token[line];
			if (lineData == null || lineData.length() == 0) {
				showMessage(source +
					" nothing in buffer");
				return;
			}
			tilesValue = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			if (tilesValue >= BarrelInterface.MIN_TILES) {
				for (i = this.tiles; i < tilesValue; i++) {
					((BarrelFrame) frame).callback(
						BarrelInterface.ACTION_INCX);
				}
				for (i = this.tiles; i > tilesValue; i--) {
					((BarrelFrame) frame).callback(
						BarrelInterface.ACTION_DECX);
				}
			} else {
				showMessage(source +
					" corrupted: tiles " + tilesValue +
					" should be at least " +
					BarrelInterface.MIN_TILES);
				return;
			}
			lineData = token[++line];
			facesValue = Integer.parseInt(lineData.substring(
					lineData.indexOf(symbol) + 2));
			if (facesValue >= BarrelInterface.MIN_FACES &&
					facesValue <= BarrelInterface.MAX_FACES) {
				for (i = this.faces; i < facesValue; i++) {
					((BarrelFrame) frame).callback(
						BarrelInterface.ACTION_INCY);
				}
				for (i = this.faces; i > facesValue; i--) {
					((BarrelFrame) frame).callback(
						BarrelInterface.ACTION_DECY);
				}
			} else {
				showMessage(source +
					" corrupted: faces " + facesValue +
					" should be at between " +
					BarrelInterface.MIN_FACES +
					" and " + BarrelInterface.MAX_FACES +
					" inclusive");
				return;
			}
			lineData = token[++line];
			pairsValue = (Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2)) != 0);
			if (this.pairs != pairsValue)
				((BarrelFrame) frame).callback(
					BarrelInterface.ACTION_PAIRS);
			lineData = token[++line];
			orientValue = (Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2)) != 0);
			if (this.orient != orientValue)
				((BarrelFrame) frame).callback(
					BarrelInterface.ACTION_ORIENTIZE);
			lineData = token[++line];
			practiceValue = (Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2)) != 0);
			if (this.practice != practiceValue)
				((BarrelFrame) frame).callback(
					BarrelInterface.ACTION_PRACTICE);
			lineData = token[++line];
			moves = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			line = scanStartPosition(++line);
			((BarrelFrame) frame).callback(
				BarrelInterface.ACTION_RESTORE);
			setStartPosition();
			allTilesPaint = true;
			repaint();
			scanMoves(++line, moves);
			System.out.println(source + ": tiles " + tilesValue +
				", faces " + facesValue + ", pairs " + pairsValue +
				", orient " + orientValue +
				", practice " + practiceValue + ", moves " + moves);
			cheat = true;
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			showMessage("Corrupt input found when reading");
		} catch (Exception e) {
			//e.printStackTrace();
			showMessage("Bad input file");
		}
	}

	public void printPuzzle(StringBuilder sb) {
		String symbol = SYMBOL;
		String space = " ";
		String newline = NL;

		sb.append("tiles").append(symbol).append(space)
			.append(tiles).append(newline);
		sb.append("faces").append(symbol).append(space)
			.append(faces).append(newline);
		sb.append("pairs").append(symbol).append(space)
			.append((pairs) ? 1 : 0).append(newline);
		sb.append("orient").append(symbol).append(space)
			.append((orient) ? 1 : 0).append(newline);
		sb.append("practice").append(symbol).append(space)
			.append((practice) ? 1 : 0).append(newline);
		sb.append("moves").append(symbol).append(space)
			.append(undo.numMoves()).append(newline);
		printStartPosition(sb);
		printMoves(sb);
	}

	public void writePuzzle() {
		StringBuilder sb = new StringBuilder();

		printPuzzle(sb);
		((BarrelFrame) frame).writeTextArea(sb.toString());
	}

	public void writePuzzleFile() {
		StringBuilder sb = new StringBuilder();

		printPuzzle(sb);
		try (PrintWriter pr = new PrintWriter(new FileWriter(
				DATAFILE, false))) {
			pr.print(sb);
			System.out.println("Saved to " + DATAFILE);
		} catch (Exception e) {
			//e.printStackTrace();
			showMessage("Cannot write to " + DATAFILE);
		}
	}

	void undoPuzzle() {
		if (mouseDown || stackPaint || randomPaints != 0 ||
				!undo.madeMoves())
			return;
		undo.getMove();
		currentDirection = undo.readMoveDirection();
		currentTile = undo.readmoveTile();
		currentShift = undo.readmoveShift();
		redo.setMove(currentDirection, currentTile, currentShift);
		// move in opposite direction
		currentDirection = (currentDirection + COORD / 2) % COORD;
		if (currentShift == 1 && (currentDirection == TOP ||
				currentDirection == BOTTOM)) {
			stackPaint = true;
			repaint();
		} else if (checkmoveTilesDir(currentDirection)) {
			stackPaint = true;
			repaint();
			started = true;
			((BarrelFrame) frame).callback(
				BarrelInterface.ACTION_UNDO);
		} else {
			currentDirection = currentShift = currentTile = -1;
		}
	}

	void redoPuzzle() {
		if (mouseDown || stackPaint || randomPaints != 0 ||
				!redo.madeMoves())
			return;
		redo.getMove();
		currentDirection = redo.readMoveDirection();
		currentTile = redo.readmoveTile();
		currentShift = redo.readmoveShift();
		undo.setMove(currentDirection, currentTile, currentShift);
		if (currentShift == 1 && (currentDirection == TOP ||
				currentDirection == BOTTOM)) {
			stackPaint = true;
			repaint();
		} else if (checkmoveTilesDir(currentDirection)) {
			stackPaint = true;
			repaint();
			started = true;
			((BarrelFrame) frame).callback(
				BarrelInterface.ACTION_REDO);
		} else {
			currentDirection = currentShift = currentTile = -1;
		}
	}

	void solvePuzzle() {
		boolean debug = false;

		if (mouseDown || stackPaint || randomPaints != 0)
			return;
		/* if (tiles == 6 && faces == 5) */
		if (!orient || debug) {
			cheat = true;
			solve.setFlag(true);
			try {
				Thread.sleep(200);
			} catch (InterruptedException e) {
				System.err.println("Interrupted");
			}
			//solve.solveSomeTiles();
		} else {
			showMessage("Auto-solver: sorry, only implemented when" +
				NL + "oriented is off.");
		}
	}

	public void numberPuzzle() {
		number = !number;
		if (orient) {
			resizePaint = true;
			repaint();
		} else
			((BarrelFrame) frame).callback(
				BarrelInterface.ACTION_ORIENTIZE);
	}

	void orientizePuzzle() {
		if (mouseDown || randomPaints != 0 || !number)
			return;
		((BarrelFrame) frame).callback(
			BarrelInterface.ACTION_ORIENTIZE);
	}

	void pairsPuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		((BarrelFrame) frame).callback(
			BarrelInterface.ACTION_PAIRS);
	}

	void practicePuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		practiceTiles();
	}

	void practicePuzzleWithQuery() {
		if (!started)
			practicePuzzle();
	}

	void printPositions() {
		for (int i = 0; i < tileFaces; i++) {
			int pos = startPosition[i];
			if (pos < 100)
				System.out.print(" ");
			if (pos < 10 && pos >= 0)
				System.out.print(" ");
			System.out.print(pos);
			if (i % tiles == tiles - 1) {
				System.out.println();
			} else {
				System.out.print(" ");
			}
		}
		System.out.println();
	}

	void getStartPosition() {
		int i;

		for (i = 0; i < PLUNGERS_NUMBER; i++)
			startSpace[i] = spacePosition[i];
		for (i = 0; i < tileFaces; i++)
			startPosition[i] = tileOfPosition[i];
	}

	void setStartPosition() {
		int i;

		for (i = 0; i < PLUNGERS_NUMBER; i++)
			spacePosition[i] = startSpace[i];
		for (i = 0; i < tileFaces; i++)
			tileOfPosition[i] = startPosition[i];
		// drawAll
	}

	int scanStartPosition(int location) {
		String lineData;
		int line = location;

		try {
			String space = "\\s+";
			int n = 0;

			++line; //read startingPosition:
			for (int j = 0; j < faces; j++) {
				line++;
				if (line >= token.length) {
					System.out.println("Short data in StartPosition");
					return -1;
				}
				lineData = token[line];
				String[] st = lineData.split(space);
				int offset = ("".equals(st[0])) ? 1 : 0;
				for (int i = 0; i < tiles; i++) {
					int num = Integer.parseInt(st[i + offset].trim());

					startPosition[n] = num;
					if (num < 0)
						startSpace[num + PLUNGERS_NUMBER] =
							n;
					n++;
				}
			}
			//blank line read
		} catch (NumberFormatException e) {
			e.printStackTrace();
			showMessage("Corrupt input in StartPosition");
			return -1;
		}
		return line;
	}

	boolean scanMoves(int location, int moves) {
		Graphics g = getGraphics();

		if (g == null)
			return false;
		try {
			int l;
			int line = location;
			String lineData;
			String symbol = SYMBOL;
			String space = "\\s+";

			++line; //read moves direction
			paint(g);
			for (l = 0; l < moves && ++line < token.length; l++) {
				lineData = token[line];
				String[] st = lineData.substring(
					lineData.indexOf(symbol) + 2).split(space);
				currentDirection = Integer.parseInt(st[0].trim());
				currentTile = Integer.parseInt(st[1].trim());
				currentShift = Integer.parseInt(st[2].trim());
				if (checkmoveTilesDir(currentDirection)) {
					scanPaint = true;
					undo.setMove(currentDirection,
						currentTile, currentShift);
					redo.flushMoves();
					if (currentShift == 0) {
						((BarrelFrame) frame).callback(
						BarrelInterface.ACTION_MOVED);
					}
					paint(g);
				} else {
					currentDirection = -1;
					System.out.println(l +
						": move in direction " +
						currentDirection +
						", cannot be made.");
					return false;
				}

			}
			if (l < moves) {
				showMessage("Corrupt input scanMoves, only " + l + " moves");
				return false;
			}
		} catch (NumberFormatException e) {
			e.printStackTrace();
			showMessage("Corrupt input in scanMoves");
			return false;
		} finally {
			g.dispose();
		}
		return true;
	}

	void printStartPosition(StringBuilder sb) {
		String symbol = SYMBOL;
		String space = " ";
		String newline = NL;

		sb.append(newline);
		sb.append("startingPosition").append(symbol).append(newline);
		for (int i = 0; i < tileFaces; i++) {
			int pos = startPosition[i];
			if (pos < 100)
				sb.append(space);
			if (pos < 10 && pos >= 0)
				sb.append(space);
			sb.append(space).append(pos);
			if (((i + 1) % tiles) == 0) {
				sb.append(newline);
			}
		}
		sb.append(newline);
	}

	void printMoves(StringBuilder sb) {
		int direction, tile, shift;
		int nMoves = undo.numMoves();
		String symbol = SYMBOL;
		String space = "\t";
		String newline = NL;

		sb.append("moves").append(space)
			.append("dir").append(space)
			.append("tile").append(space)
			.append("shift").append(newline);
		while (undo.madeMoves()) {
			undo.getMove();
			direction = undo.readMoveDirection();
			shift = undo.readmoveShift();
			tile = undo.readmoveTile();
			redo.setMove(direction, tile, shift);
		}
		// apply all moves back to the stack since
		// we emptied it above!!
		for (int i = 0; i < nMoves; i++) {
			redo.getMove();
			direction = redo.readMoveDirection();
			shift = redo.readmoveShift();
			tile = redo.readmoveTile();
			sb.append(i + 1).append(symbol).append(space)
				.append(direction).append(space)
				.append(tile).append(space)
				.append(shift).append(newline);
			undo.setMove(direction, tile, shift);
		}
	}

	@Override
	public void paint(Graphics g) {
		if (firstPaint) {
			initializePuzzle();
			firstPaint = false;
			resizePaint = true;
		}
		if (!framePaint && !allTilesPaint &&
				!selectPaint && !motionPaint &&
				!releasePaint && !movePaint &&
				!scanPaint && !stackPaint &&
				randomPaints == 0) {
			resizePaint = true;
		}
		if (resizePaint) {
			resizePuzzle();
			exposePuzzle();
			resizePaint = false;
		}
		if (framePaint) {
			drawFrame(true);
			drawFrame(bufferedFrameGraphics, true);
			framePaint = false;
		}
		if (allTilesPaint) {
			drawAllTiles();
			allTilesPaint = false;
		}
		if (selectPaint) {
			if (currentPosition != -1) {
				drawTile(currentPosition,
					(tileOfPosition[currentPosition] < 0),
					false, 1, 0, 0);
			}
			selectPaint = false;
		}
		if (motionPaint) {
			if (currentPosition != -1) {
				int j = nextPosition / tiles;
				int k = currentPosition;

				drawTile(k, true, true, 1, 0, 0);
				drawTile(k, tileOfPosition[k] <= 0,
					tileOfPosition[k] <= 0, 0, 0, 0);
				diffRecord(motionDiff, j, false);
				diffRecord(shiftDiff, j, true);
				if (nextShift == 0 && checkSolved()) {
					((BarrelFrame) frame).callback(
						BarrelInterface.ACTION_SOLVED);
				}
				if (tileOfPosition[currentPosition] > 0) {
					drawTile(currentPosition,
						false, false, 0, 0, 0);
				}
				motionDiff = 0;
				shiftDiff = 0;
				if (currentSpace != -1) {
					motionTiles();
				}
				currentPosition = nextPosition;
			}
			motionPaint = false;
		}
		if (releasePaint) {
			if (currentPosition != -1) {
				drawTile(currentPosition,
					tileOfPosition[currentPosition] < 0,
					true, 1, 0, 0);
				if (tileOfPosition[currentPosition] > 0) {
					drawTile(currentPosition,
						false, false, 0, 0, 0);
				}
				if (currentSpace != -1) {
					selectTiles();
				}
				currentPosition = -1;
			}
			releasePaint = false;
		}
		if (movePaint || scanPaint || stackPaint) {
			int fast;

			if (stackPaint)
				fast = DOUBLE;
			else if (movePaint)
				fast = NORMAL;
			else
				fast = INSTANT;
			if (currentShift == 1 && (currentDirection == TOP ||
					currentDirection == BOTTOM)) {
				moveShiftCb(currentDirection, fast);
			} else {
				moveTilesDir(currentDirection,
					currentTile, false, fast);
				if (stackPaint && checkSolved()) {
					((BarrelFrame) frame).callback(
						BarrelInterface.ACTION_SOLVED);
				}
				started = true;
			}
			currentDirection = currentShift = currentTile = -1;
			movePaint = false;
			scanPaint = false;
			stackPaint = false;
		}
		if (randomPaints > 0) {
			randomizingTiles();
		}
	}

	@Override
	public void update(Graphics g) {
		paint(g); /* no erase */
	}
}
